/* (C) 2022 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
 * Author: Oliver Smith <osmith@sysmocom.de>
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/lienses/>.
 *
 */

#include "config.h"
#include <osmocom/core/msgb.h>
#include <osmocom/gsm/gsm0808.h>
#include <osmocom/sigtran/sccp_helpers.h>
#include <osmocom/bsc_nat/logging.h>
#include <osmocom/bsc_nat/subscr_conn.h>
#include <osmocom/bsc_nat/subscr_conn_fsm.h>

static int bssap_conn_cn_fwd(struct subscr_conn *subscr_conn, struct msgb *msg, unsigned int length)
{
	LOGP(DMAIN, LOGL_DEBUG, "Fwd BSSAP via %s\n", talloc_get_name(subscr_conn));

	msgb_pull_to_l2(msg);

	return osmo_sccp_tx_data(g_bsc_nat->ran.sccp_inst->scu,
				 subscr_conn->ran.id,
				 msg->data,
				 msgb_length(msg));
}

static int bssmap_cn_handle_assignment_rqst(struct subscr_conn *subscr_conn, struct msgb *msg, unsigned int length)
{
	int rc;
	unsigned int rc_u;
	struct tlv_parsed tp;
	uint8_t cause;
	struct sockaddr_storage rtp_addr;
	struct gsm0808_channel_type ct;
	struct bssmap_assignment_rqst *ass_req = &subscr_conn->cn.ass_req;

	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l3h + 1, length - 1, 0, 0);

	/* Get channel type */
	if (!TLVP_PRESENT(&tp, GSM0808_IE_CHANNEL_TYPE)) {
		LOGP(DMAIN, LOGL_ERROR, "Mandatory channel type not present\n");
		cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
		goto reject;
	}
	rc = gsm0808_dec_channel_type(&ct,  TLVP_VAL(&tp, GSM0808_IE_CHANNEL_TYPE),
				      TLVP_LEN(&tp, GSM0808_IE_CHANNEL_TYPE));
	if (rc < 0) {
		LOGP(DMAIN, LOGL_ERROR, "Unable to decode channel type\n");
		cause = GSM0808_CAUSE_INCORRECT_VALUE;
		goto reject;
	}

	/* Not speech: fwd directly */
	if ((ct.ch_indctr & 0x0f) != GSM0808_CHAN_SPEECH) {
		LOGP(DMAIN, LOGL_DEBUG, "Channel type is not speech, forwarding directly\n");
		rc = bssap_conn_cn_fwd(subscr_conn, msg, length);
		if (rc < 0) {
			cause = GSM0808_CAUSE_PROTOCOL_ERROR_BETWEEN_BSS_AND_MSC;
			goto reject;
		}
		return 0;
	}

	/* Get AoIP transport address */
	if (!TLVP_PRESENT(&tp, GSM0808_IE_AOIP_TRASP_ADDR)) {
		LOGP(DMAIN, LOGL_ERROR, "AoIP transport address missing in ASSIGN REQ\n");
		cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
		goto reject;
	}
	rc = gsm0808_dec_aoip_trasp_addr(&rtp_addr,
					 TLVP_VAL(&tp, GSM0808_IE_AOIP_TRASP_ADDR),
					 TLVP_LEN(&tp, GSM0808_IE_AOIP_TRASP_ADDR));
	if (rc < 0) {
		LOGP(DMAIN, LOGL_ERROR, "Unable to decode AoIP transport address\n");
		cause = GSM0808_CAUSE_INCORRECT_VALUE;
		goto reject;
	}

	rc_u = osmo_sockaddr_to_str_and_uint(ass_req->rtp_addr, sizeof(ass_req->rtp_addr), &ass_req->rtp_port,
					     (const struct sockaddr *)&rtp_addr);
	if (!rc_u || rc_u >= sizeof(ass_req->rtp_addr)) {
		LOGP(DMAIN, LOGL_ERROR, "Assignment request: RTP address is too long\n");
		cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
		goto reject;
	}

	/* Get more IEs? */

	/* Don't forward the message directly. Instead, let the subscr_conn FSM
	 * allocate new MGCP connections in the BSCNAT's MGW and then send a
	 * similar assignment request, but with the RTP address replaced. */
	if (osmo_fsm_inst_dispatch(subscr_conn->fi, SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_REQUEST, NULL) < 0) {
		cause = GSM0808_CAUSE_PROTOCOL_ERROR_BETWEEN_BSS_AND_MSC;
		goto reject;
	}

	return 0;
reject:
	/* FIXME: send error */
	LOGP(DMAIN, LOGL_ERROR, "STUB: send error cause %d\n", cause);
	return -1;
}

static int bssmap_conn_cn_rcvmsg_dt(struct subscr_conn *subscr_conn, struct msgb *msg, unsigned int length)
{
	int ret = 0;

	switch (msg->l3h[0]) {
	case BSS_MAP_MSG_ASSIGNMENT_RQST:
		ret = bssmap_cn_handle_assignment_rqst(subscr_conn, msg, length);
		break;
	default:
		ret = bssap_conn_cn_fwd(subscr_conn, msg, length);
		break;
	}

	return ret;
}

static int bssmap_conn_rcvmsg_dt(enum bsc_nat_net net, struct subscr_conn *subscr_conn, struct msgb *msg,
				  unsigned int length)
{
	if (length < 1) {
		LOGP(DMAIN, LOGL_ERROR, "Not enough room: %u\n", length);
		return -1;
	}

	LOGP(DMAIN, LOGL_NOTICE, "Rx DT BSSMAP %s\n", gsm0808_bssmap_name(msg->l3h[0]));

	if (net == BSC_NAT_NET_CN)
		return bssmap_conn_cn_rcvmsg_dt(subscr_conn, msg, length);

	LOGP(DMAIN, LOGL_ERROR, "%s(RAN) is not implemented!\n", __func__);
	return -1;
}

int bssap_conn_handle_dt(enum bsc_nat_net net, struct subscr_conn *subscr_conn, struct msgb *msgb, unsigned int length)
{
	struct bssmap_header *bs;
	int rc = -1;

	LOGP(DMAIN, LOGL_DEBUG, "Rx DT: %s\n", osmo_hexdump(msgb->l2h, length));

	if (length < sizeof(*bs)) {
		LOGP(DMAIN, LOGL_ERROR, "The header is too short\n");
		return -1;
	}

	switch (msgb->l2h[0]) {
	case BSSAP_MSG_BSS_MANAGEMENT:
		bs = (struct bssmap_header *)msgb->l2h;
		if (bs->length < length - sizeof(*bs)) {
			LOGP(DMAIN, LOGL_ERROR, "Failed to parse BSSMAP header\n");
			return -1;
		}
		msgb->l3h = &msgb->l2h[sizeof(*bs)];
		rc = bssmap_conn_rcvmsg_dt(net, subscr_conn, msgb, length - sizeof(*bs));
		break;
	case BSSAP_MSG_DTAP:
		LOGP(DMAIN, LOGL_DEBUG, "Rx DT DTAP\n");
		if (net == BSC_NAT_NET_CN) {
			rc = bssap_conn_cn_fwd(subscr_conn, msgb, length);
		} else {
			LOGP(DMAIN, LOGL_ERROR, "%s(RAN) is not implemented!\n", __func__);
			rc = -1;
		}
		break;
	default:
		LOGP(DMAIN, LOGL_ERROR, "%s(%s) is not implemented!\n", __func__, gsm0808_bssap_name(msgb->l2h[0]));
	}

	return rc;
}
