/* (C) 2022 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
 * Author: Oliver Smith <osmith@sysmocom.de>
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/lienses/>.
 *
 */

#include "config.h"
#include <inttypes.h>
#include <osmocom/core/fsm.h>
#include <osmocom/mgcp_client/mgcp_client_endpoint_fsm.h>
#include <osmocom/mgcp_client/mgcp_client_fsm.h>
#include <osmocom/bsc_nat/bsc_nat.h>
#include <osmocom/bsc_nat/bssap.h>
#include <osmocom/bsc_nat/logging.h>
#include <osmocom/bsc_nat/subscr_conn.h>
#include <osmocom/bsc_nat/subscr_conn_fsm.h>

#define X(s) (1 << (s))

static void st_idle(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_REQUEST:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_processing_assignment_request_on_enter(struct osmo_fsm_inst *fi, uint32_t prev_stat)
{
	struct mgcp_client *mgcp_client;
	struct mgcp_conn_peer crcx_info = {};
	struct subscr_conn *subscr_conn = (struct subscr_conn *)fi->priv;
	struct bssmap_assignment_rqst *ass_req = &subscr_conn->cn.ass_req;

	/* MGCP EP */
	mgcp_client = mgcp_client_pool_get(g_bsc_nat->mgw.pool);
	OSMO_ASSERT(mgcp_client);
	subscr_conn->ep = osmo_mgcpc_ep_alloc(subscr_conn->fi, SUBSCR_CONN_FSM_EV_MGCP_EP_TERM, mgcp_client,
					      g_bsc_nat->mgw.tdefs, "SUBSCR-CONN-EP",
					      mgcp_client_rtpbridge_wildcard(mgcp_client));

	crcx_info.call_id = subscr_conn->ran.id; /* FIXME: generate unique call id and store in subscr_conn */

	/* MGCP CRCX RAN */
	subscr_conn->ran.ci = osmo_mgcpc_ep_ci_add(subscr_conn->ep, "CI(%s:%"PRIu32")",
						   talloc_get_name(subscr_conn->ran.bsc), subscr_conn->ran.id);
	osmo_mgcpc_ep_ci_request(subscr_conn->ran.ci, MGCP_VERB_CRCX, &crcx_info, subscr_conn->fi,
				 SUBSCR_CONN_FSM_EV_MGCP_EP_OK, SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL, NULL);

	/* Set RTP addr + port */
	osmo_static_assert(sizeof(crcx_info.addr) == sizeof(ass_req->rtp_addr), sizeof_rtp_addr);
	memcpy(crcx_info.addr, ass_req->rtp_addr, sizeof(crcx_info.addr));
	crcx_info.port = ass_req->rtp_port;

	/* MGCP CRCX CN */
	subscr_conn->cn.ci = osmo_mgcpc_ep_ci_add(subscr_conn->ep, "CI(%s:%"PRIu32")",
						  talloc_get_name(subscr_conn->cn.msc), subscr_conn->cn.id);
	osmo_mgcpc_ep_ci_request(subscr_conn->cn.ci, MGCP_VERB_CRCX, &crcx_info, subscr_conn->fi,
				 SUBSCR_CONN_FSM_EV_MGCP_EP_OK, SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL, NULL);
	return;
}

static void st_processing_assignment_request(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_MGCP_EP_OK:
		LOGP(DMAIN, LOGL_DEBUG, "Rx MGCP EP OK 1/2\n");
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST_WAIT_EP2, 0, 0);
		break;
	case SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL:
		LOGP(DMAIN, LOGL_ERROR, "MGCP failure, aborting assignment request processing\n");
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_IDLE, 0, 0);
		/* FIXME: do mgcp abort? send bssmap error? */
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_processing_assignment_request_wait_ep2(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_MGCP_EP_OK:
		LOGP(DMAIN, LOGL_DEBUG, "Rx MGCP EP OK 2/2\n");
		/* FIXME: send bssmap assignment request with changed rtp address to bsc */
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_WAITING_FOR_ASSIGNMENT_COMPLETE, 0, 0);
		break;
	case SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL:
		LOGP(DMAIN, LOGL_ERROR, "MGCP failure, aborting assignment request processing\n");
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_IDLE, 0, 0);
		/* FIXME: do mgcp abort? send bssmap error? */
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_waiting_for_assignment_complete(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_COMPLETE:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_COMPLETE, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_processing_assignment_complete(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_MGCP_EP_OK:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMMAND, 0, 0);
		break;
	case SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL:
		LOGP(DMAIN, LOGL_ERROR, "MGCP failure, aborting assignment complete processing\n");
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_IDLE, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_waiting_for_clear_command(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMMAND:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_PROCESSING_CLEAR_COMMAND, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_processing_clear_command(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_MGCP_EP_OK:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMPLETE, 0, 0);
		break;
	case SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL:
		LOGP(DMAIN, LOGL_ERROR, "MGCP failure, aborting clear command processing\n");
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_IDLE, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static void st_waiting_for_clear_complete(struct osmo_fsm_inst *fi, uint32_t event, void *data)
{
	switch (event) {
	case SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMPLETE:
		osmo_fsm_inst_state_chg(fi, SUBSCR_CONN_FSM_ST_IDLE, 0, 0);
		break;
	default:
		OSMO_ASSERT(false);
	}
}

static struct osmo_fsm_state subscr_conn_fsm_states[] = {
	[SUBSCR_CONN_FSM_ST_IDLE] = {
		.name = "IDLE",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_REQUEST)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST)
			,
		.action = st_idle,
	},
	[SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST] = {
		.name = "PROCESSING_ASSIGNMENT_REQUEST",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_OK)
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_IDLE)
			| X(SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST_WAIT_EP2)
			,
		.action = st_processing_assignment_request,
		.onenter = st_processing_assignment_request_on_enter,
	},
	[SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_REQUEST_WAIT_EP2] = {
		.name = "PROCESSING_ASSIGNMENT_REQUEST_WAIT_EP2",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_OK)
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_IDLE)
			| X(SUBSCR_CONN_FSM_ST_WAITING_FOR_ASSIGNMENT_COMPLETE)
			,
		.action = st_processing_assignment_request_wait_ep2,
	},
	[SUBSCR_CONN_FSM_ST_WAITING_FOR_ASSIGNMENT_COMPLETE] = {
		.name = "WAITING_FOR_ASSIGNMENT_COMPLETE",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_COMPLETE)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_COMPLETE)
			,
		.action = st_waiting_for_assignment_complete,
	},
	[SUBSCR_CONN_FSM_ST_PROCESSING_ASSIGNMENT_COMPLETE] = {
		.name = "PROCESSING_ASSIGNMENT_COMPLETE",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_OK)
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_IDLE)
			| X(SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMMAND)
			,
		.action = st_processing_assignment_complete,
	},
	[SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMMAND] = {
		.name = "WAITING_FOR_CLEAR_COMMAND",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMMAND)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_PROCESSING_CLEAR_COMMAND)
			,
		.action = st_waiting_for_clear_command,
	},
	[SUBSCR_CONN_FSM_ST_PROCESSING_CLEAR_COMMAND] = {
		.name = "PROCESSING_CLEAR_COMMAND",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_OK)
			| X(SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_IDLE)
			| X(SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMPLETE)
			,
		.action = st_processing_clear_command,
	},
	[SUBSCR_CONN_FSM_ST_WAITING_FOR_CLEAR_COMPLETE] = {
		.name = "WAITING_FOR_CLEAR_COMPLETE",
		.in_event_mask = 0
			| X(SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMPLETE)
			,
		.out_state_mask = 0
			| X(SUBSCR_CONN_FSM_ST_IDLE)
			,
		.action = st_waiting_for_clear_complete,
	},
};

const struct value_string subscr_conn_fsm_event_names[] = {
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_REQUEST),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_BSSMAP_ASSIGNMENT_COMPLETE),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMMAND),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_BSSMAP_CLEAR_COMPLETE),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_MGCP_EP_OK),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_MGCP_EP_FAIL),
	OSMO_VALUE_STRING(SUBSCR_CONN_FSM_EV_MGCP_EP_TERM),
	{ 0, NULL }
};

struct osmo_fsm subscr_conn_fsm = {
	.name = "SUBSCR_CONN",
	.states = subscr_conn_fsm_states,
	.num_states = ARRAY_SIZE(subscr_conn_fsm_states),
	.log_subsys = DMAIN,
	.event_names = subscr_conn_fsm_event_names,
};

static __attribute__((constructor)) void subscr_conn_fsm_init(void)
{
	OSMO_ASSERT(osmo_fsm_register(&subscr_conn_fsm) == 0);
}
